//
//	Copyright (c) 2009 Nemetschek North America 
//	All Rights Reserved.
//
// Created by Alexandra Nikolova

#pragma once

namespace VWFC
{
	namespace Tools
	{
		class CRijndael;

		// ------------------------------------------------------------------------------------------------------
		class CPassword
		{
		public:
						CPassword();
						CPassword(const void* memory, size_t size);
						CPassword(const CPassword& src);
			virtual		~CPassword();

			CPassword&	operator=(const CPassword& src);
			Sint8		operator[](const size_t index) const;
			Sint8&		operator[](const size_t index);

			size_t		size() const;
			void		reserve(size_t size);
			void*		memory() const;
			void		clear();
			
			size_t		GetLettersCount() const;
			void		SetLettersCount(size_t cnt);
			bool		IsEqualTo(const CPassword& pass, const TXString& key) const;
		
		protected:
			void		Init(size_t blockSize);

		protected:
			friend class CRijndael;

		private:
			size_t		fLettersCnt;
			size_t		fDataBlockSize;
			Sint8*		fData;
		};

		//Rijndael (pronounced Reindaal) is a block cipher, designed by Joan Daemen and Vincent Rijmen as a candidate algorithm for the AES.
		//The cipher has a variable block length and key length. The authors currently specify how to use keys with a length
		//of 128, 192, or 256 bits to encrypt blocks with al length of 128, 192 or 256 bits (all nine combinations of
		//key length and block length are possible). Both block length and key length can be extended very easily to
		// multiples of 32 bits.

		// This implementation is based on http://www.codeproject.com/KB/security/aes.aspx

		//Operation Modes
		// - In the Electronic Code Book (eRijndaelModes_ECB) mode if the same block is encrypted twice with the same key, the resulting
		//ciphertext blocks are the same.
		// - In the Cipher Block Chaining (eRijndaelModes_CBC) mode a ciphertext block is obtained by first xoring the
		//plaintext block with the previous ciphertext block, and encrypting the resulting value.
		// - In Cipher Feedback Block (eRijndaelModes_CFB) mode a ciphertext block is obtained by encrypting the previous ciphertext block
		//and xoring the resulting value with the plaintext.

		// ------------------------------------------------------------------------------------------------------
		enum ERijndaelModes 
		{
			eRijndaelModes_ECB,
			eRijndaelModes_CBC,
			eRijndaelModes_CFB
		};

		enum ERijndaelErrors
		{
			eRijndaelErrors_NoError,
			eRijndaelErrors_ObjectNotInitialized,
			eRijndaelErrors_DataNotMultipleOfBlockSize,
			eRijndaelErrors_Emptykey,
			eRijndaelErrors_IncorrectKeyLength,
			eRijndaelErrors_IncorrectBlockLength,
			eRijndaelErrors_NothingToEncrypt,
			eRijndaelErrors_NothingToDecrypt
		};

		// ------------------------------------------------------------------------------------------------------
		const size_t kKeyLength8		= 8;
		const size_t kKeyLength16		= 16;
		const size_t kKeyLength24		= 24;
		const size_t kKeyLength32		= 32;
		
		const size_t kBlockSize8		= 8;
		const size_t kBlockSize16		= 16;
		const size_t kBlockSize24		= 24;
		const size_t kBlockSize32		= 32;
		const size_t kBlockSize128		= 128;
		const size_t kBlockSize192		= 192;
		const size_t kBlockSize256		= 256;

		const size_t kRounds10			= 10;
		const size_t kRounds12			= 12;
		const size_t kRounds14			= 14;

		const size_t kMaxKC				= 8;
		const size_t kMaxBC				= 8;

		// ------------------------------------------------------------------------------------------------------
		class CRijndael
		{
		public:	
								CRijndael();
			virtual				~CRijndael();

			ERijndaelErrors		MakeKey(const TXString& key);
			
			ERijndaelErrors		Encrypt(const TXString& inText, CPassword& outPass, ERijndaelModes iMode = eRijndaelModes_ECB);
			ERijndaelErrors		Decrypt(const CPassword& inPass, TXString& outText, ERijndaelModes iMode = eRijndaelModes_ECB);

			bool				AreEqual(const CPassword& passA, const CPassword& passB);

		protected:
			ERijndaelErrors		EncryptBlock(Sint8 const* blockIn, Sint8*& blockOut);
			ERijndaelErrors		DecryptBlock(Sint8 const* blockIn, Sint8*& blockOut);
			ERijndaelErrors		Encrypt(Sint8 const* blockIn, Sint8* blockOut, size_t n, ERijndaelModes iMode = eRijndaelModes_ECB);
			ERijndaelErrors		Decrypt(Sint8 const* blockIn, Sint8* blockOut, size_t n, ERijndaelModes iMode = eRijndaelModes_ECB);
			
			inline int			GetKeyLength();
			inline int			GetBlockSize();
			inline int			GetRounds();
			inline void			ResetChain();

			void				BlockToString(Sint8*& block, TXString& outString, size_t lettersCnt);

		protected:
			ERijndaelErrors		Xor(Sint8* buff, Sint8 const* chain);
			ERijndaelErrors		DefEncryptBlock(Sint8 const* blockIn, Sint8*& blockOut);
			ERijndaelErrors		DefDecryptBlock(Sint8 const* blockIn, Sint8*& blockOut);

		private:
			static Sint8 const*		farrChainNULL;				// Null chain
			static const int		farrAlog[kBlockSize256];
			static const int		farrLog[kBlockSize256];
			static const Sint8		farrS[kBlockSize256];
			static const Sint8		farrSi[kBlockSize256];
			static const int		farrT1[kBlockSize256];
			static const int		farrT2[kBlockSize256];
			static const int		farrT3[kBlockSize256];
			static const int		farrT4[kBlockSize256];
			static const int		farrT5[kBlockSize256];
			static const int		farrT6[kBlockSize256];
			static const int		farrT7[kBlockSize256];
			static const int		farrT8[kBlockSize256];
			static const int		farrU1[kBlockSize256];
			static const int		farrU2[kBlockSize256];
			static const int		farrU3[kBlockSize256];
			static const int		farrU4[kBlockSize256];
			static const Sint8		farrRcon[30];
			static const int		farrShiftS[3][4][2];
			
			// Key Initialization Flag
			bool					fbKeyInit;
			
			// Encryption (fEncryptKeys) round key
			int**					fEncryptKeys/*[kRounds14 + 1][kMaxBC]*/;

			// Decryption (fDecryptKeys) round key
			int**					fDecryptKeys/*[kRounds14 + 1][kMaxBC]*/;
			
			// Key Length
			size_t					fnKeylength;
			
			// Block Size
			size_t					fnBlockSize;
			
			// Number of Rounds
			size_t					fniRounds;
			
			// Chain Block
			Sint8					farrChain0[kBlockSize32];
			Sint8					farrChain[kBlockSize32];
		};
		
		// ------------------------------------------------------------------------
		// Inlines
		inline int CRijndael::GetKeyLength()
		{
			if ( false == fbKeyInit )
				return -1;
			else
				return fnKeylength;
		}

		inline int CRijndael::GetBlockSize()
		{
			if ( false == fbKeyInit )
				return -1;
			return fnBlockSize;
		}

		//Number of Rounds
		inline int CRijndael::GetRounds()
		{
			if ( false == fbKeyInit)
				return -1;
			else
				return fniRounds;
		}

		inline void CRijndael::ResetChain()
		{
			memcpy(farrChain, farrChain0, fnBlockSize);
		}
	}
}
